extends KinematicBody

export(NodePath) onready var body = get_node(body) as Spatial
export(NodePath) onready var head = get_node(head) as Spatial
export(NodePath) onready var cam = get_node(cam) as Camera
#onready var col_stand = $CollisionStand
#onready var col_crouch = $CollisionCrouch
export(NodePath) onready var tween = get_node(tween) as Tween

var mouse_sensitivity := 10.0

var stop_speed := 4.0
var move_speed := 10.0
var max_slope_angle := 45

var gravity := 60.0
var max_fall_speed := 50.0
var jump_height := 2.0

var accel_ground := 10.0
var accel_air := 3.0
var accel_water := 4.0

var friction_ground := 6.0
var friction_water := 1.0

var step_height := 0.5

var queue_jump := false
var snap_vector := Vector3.ZERO
var grounded := false
var wall := false
var crouch := false
var on_ladder := false
var swimming := false

const cr_sp = 0.3
const ldr_sp = 0.5

onready var head_offset := [head.transform.origin.y,head.transform.origin.z]

#onready var ray_rotate = $RayRotate
#onready var rays = [$RayRotate/StepForward, $RayRotate/StepR, $RayRotate/StepL, $RayRotate/StepR2, $RayRotate/StepL2, $RayRotate/StepR3, $RayRotate/StepL3, ]
#onready var rays_down = $RayRotate/StepDown
onready var collision_shape = $CollisionStand

#onready var base_of_shape = collision_shape.shape.get_height() / 2

var velocity := Vector3.ZERO

func _ready():
	Input.set_mouse_mode(Input.MOUSE_MODE_CAPTURED)
#	ray_rotate.set_as_toplevel(true)
#	rays_down.set_as_toplevel(true)
#	rays_down.set_cast_to(Vector3(0,-step_height,0))

func get_move_direction() -> Vector3:
	
	var input_dir := Vector2(
		Input.get_axis("move_forward","move_backward"),
		Input.get_axis("move_left", "move_right")
	).clamped(1.0)
	
	if !on_ladder and !swimming:
		return input_dir.x * body.global_transform.basis.z + input_dir.y * body.global_transform.basis.x
	else:
		return input_dir.x * head.global_transform.basis.z + input_dir.y * body.global_transform.basis.x

func _input(event):
	if event is InputEventMouseMotion:
		rotate_look(event.relative * 0.001 * mouse_sensitivity)
	
	if Input.is_action_just_pressed("crouch"):
		crouch()
	
	if Input.is_action_just_pressed("move_jump") and !crouch:
		queue_jump = true
	if Input.is_action_just_released("move_jump") and !crouch:
		queue_jump = false
	
	if Input.is_action_just_pressed("exit"):
		get_tree().quit()

#Kiwwi's head_bonker
#func can_uncrouch() -> bool:
#	return not test_move( Transform( collision_shape.global_transform.basis, collision_shape.global_transform.origin + Vector3.UP * 1 ), collision_shape.global_transform.origin + Vector3.UP * 1.5, false )

func crouch() -> void:
	if grounded:
		if not crouch:
#			collision_shape.shape.height = 0.5
#			translation.y -= collision_shape.shape.height * 0.5
			crouch_tween()
			crouch = true
			
		else:
			if overlap_check() == []:
#			if can_uncrouch():
#				collision_shape.shape.height = 2
#				translation.y += collision_shape.shape.height * 0.52
				crouch_tween()
				crouch = false

func crouch_tween():
	tween.stop_all()
	tween.interpolate_property(head, "translation",
		head.translation,
			Vector3( 0, head_offset[0], head_offset[1] ) if crouch else Vector3( 0, -0.75, head_offset[1]),
				0.2,
					Tween.EASE_OUT,Tween.EASE_OUT)
	tween.start()

func rotate_look(amount : Vector2) -> void:
	body.rotation.y -= amount.x
	head.rotation.x = clamp(head.rotation.x - amount.y, -PI * 0.5, PI * 0.5)

func _physics_process(delta):
	grounded = is_on_floor()
	wall = is_on_wall()
	
	velocity.y = max(-max_fall_speed, velocity.y - (gravity if (not grounded and not on_ladder) else 0.0) * delta)
	
	var h_target_dir : Vector3 = get_move_direction()
	
	if !on_ladder:
		if grounded:
			apply_friction(delta)
			accelerate(delta, h_target_dir, move_speed * (cr_sp if crouch else 1.0), accel_ground)
	#		snap_vector = -get_floor_normal()
			
			snap_vector = -get_floor_normal() - get_floor_velocity() * delta
			
			# Workaround for sliding down after jump on slope
			if velocity.y < 0:
				velocity.y = 0
				
			if queue_jump:
				velocity.y = sqrt(2 * jump_height * abs(gravity))
				queue_jump = false
				snap_vector = Vector3.ZERO
			
			else: #kiwi: not gonna step while jumping
				up_stepping(delta)
		else:
			# Workaround for 'vertical bump' when going off platform
			if snap_vector != Vector3.ZERO && velocity.y != 0:
				velocity.y = 0
				
			accelerate(delta, h_target_dir, move_speed, accel_air)
			snap_vector = Vector3.ZERO
	elif on_ladder:
		snap_vector = Vector3.ZERO
		apply_friction(delta)
		accelerate(delta, h_target_dir, move_speed * ldr_sp, accel_ground)
		if test_move(global_transform,Vector3(velocity.x,0,velocity.z),true):
			velocity.x  = 0
			
	velocity = move_and_slide_with_snap(velocity, snap_vector, Vector3.UP, true, 4, deg2rad(max_slope_angle), false)

func accelerate(delta : float, p_target_dir : Vector3, p_target_speed : float, p_accel : float):
	var current_speed : float = velocity.dot(p_target_dir)
	var add_speed : float = p_target_speed - current_speed
	#old code
	if add_speed > 0:
		var accel_speed : float = min(add_speed, p_accel * delta * p_target_speed)
		velocity += p_target_dir * accel_speed
	#new code
#	if add_speed <= 0:
#		return
#	var accel_speed : float = p_accel * delta * p_target_speed
#	if accel_speed > add_speed:
#		accel_speed = add_speed 
#	velocity.x += accel_speed * p_target_dir.x
#	velocity.z += accel_speed * p_target_dir.z

func apply_friction(delta : float):
	
#	var vec : Vector3 = velocity
	var speed : float = velocity.length()
	if is_zero_approx(speed):
		velocity = Vector3.ZERO
		return
	
	var drop : float = 0.0
	var control : float = max(speed, stop_speed)
	
	# ground friction
	if grounded or on_ladder:
		drop += control * friction_ground * delta
	
	# water friction not implemented
	var new_speed : float = max(0.0, speed - drop)
	new_speed /= speed
	
	velocity *= new_speed

#func stair_step(delta):
#	var vel_norm = velocity
#	var vel_lateral = Vector3(vel_norm.x, 0, vel_norm.z)
#	var vel_lateral_length = vel_lateral.length()
#	var vel_lateral_norm = vel_lateral.normalized()
#	if vel_lateral_length != 0 and grounded:
#		var magic_pos = Vector3.FORWARD * max(vel_lateral_length,1.7) * 2 * delta
#		ray_rotate.global_transform.origin = global_transform.origin
#		ray_rotate.look_at(global_transform.origin + vel_lateral, Vector3.UP)
#		for i in rays:
#			i.set_cast_to(magic_pos)
#			if i.is_colliding() and i.get_collision_normal().y < 0.01:
#				rays_down.global_transform.origin = i.to_global(i.get_cast_to())
#				rays_down.transform.origin += (Vector3.UP) + (vel_lateral_norm * 0.2)
#				if rays_down.is_colliding():
#					global_transform.origin.y = rays_down.get_collision_point().y + base_of_shape + 0.1
#				break

# old
func overlap_check():
	var space: PhysicsDirectSpaceState = get_world().direct_space_state as PhysicsDirectSpaceState

	var params: = PhysicsShapeQueryParameters.new()
	params.set_shape(collision_shape.shape)
	params.set_transform(global_transform)
	params.set_exclude([self])

	var castResult = space.intersect_shape(params)
	return castResult

# Kiwwi's stepping
func up_stepping( delta:float ):
	var space_state : PhysicsDirectSpaceState = get_world().direct_space_state as PhysicsDirectSpaceState
	var param := PhysicsShapeQueryParameters.new()
	var motion :Vector3 = velocity * delta
	var reminder :Vector3
	var extra_margin := 0.1
#	var current_shape = col_stand if !col_stand.disabled else col_crouch
	
	param.collision_mask = self.collision_mask
	param.set_shape( collision_shape.shape )
	param.transform = collision_shape.global_transform
	param.margin = collision_shape.shape.margin
	param.exclude = [ self ]

	var trace := space_state.cast_motion( param, motion )

	if not (trace[0] == 1.0 and trace[1] == 1.0): # frontal collision
		
		param.transform.origin += motion * trace[1]     # I think if reminder become too small, trace step down will miss the collision later. With this, change shape will collide with step but cast_motion ignore it based on document, no worries
		reminder = motion - (motion * trace[1])
		motion = Vector3.UP * (step_height + extra_margin)
		trace = space_state.cast_motion( param, motion )

		if trace[0] == 1.0 and trace[1] == 1.0: # no ceiling collision
			param.transform.origin += motion
			trace = space_state.cast_motion( param, reminder )

			if trace[0] > 0.0: # if there's space forward..
				param.transform.origin += reminder * trace[1]

				# that code_block could be replaced by a recursive function
				if not (trace[0] == 1.0 and trace[1] == 1.0):     # frontal collision with next-step → cast_up
					motion = Vector3.UP * (step_height + extra_margin)
					trace = space_state.cast_motion( param, motion )

					if trace[0] == 1.0 and trace[1] == 1.0: # no ceiling collision → cast_forward
						param.transform.origin += motion
						reminder -= reminder * trace[1]
						trace = space_state.cast_motion( param, reminder )

						if trace[0] == 1.0 and trace[1] == 1.0:
							param.transform.origin += reminder

				motion = Vector3.DOWN * (step_height + extra_margin)
				trace = space_state.cast_motion( param, motion )

				if not (trace[0] == 1.0 and trace[1] == 1.0): # collision with step-surface
					param.transform.origin += motion * trace[1]
					var collision     := space_state.get_rest_info(param)

					if not collision.empty() and not Vector3.UP.angle_to( collision.get("normal") ) > max_slope_angle:      # avoid climb "mountains"
						if not collision.get("point").y - self.global_transform.origin.y > step_height:
							self.transform.origin = param.transform.origin
							#cam.translate( Vector3.DOWN * STEP_HEIGHT )
							
							# using Zylann's plugin to visualize shape-casting
							DebugDraw.draw_box( param.transform.origin, Vector3( collision_shape.shape.radius, collision_shape.shape.height, collision_shape.shape.radius ) * 2, Color(0,1,0) )
							#is_stepping = true
	
	#is_stepping = false
